
#ifdef WIN32
#define STRICT
#include <windowsx.h>
#include <windows.h>

#else
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#ifndef Sleep		// time is in approx millisecs
#define Sleep(x) usleep( (x)<<10 )
#endif

#endif

#include <stdio.h>
#include <sys/stat.h>
#include <assert.h>
#include "hhsclnt.h"

#define boole int
#define uns8 unsigned char
#define uns16 unsigned short

#define HHS_LANGID				ENGLISH_LANGID

#define lengthof( x )	int( sizeof(x)/sizeof(x[0]) )


////////////////////////////////////////////////////////////////
// Enrollment Data

const char* prompts[] =
{
	// first, prompts that cover the english phoneme set (don't touch!)
	"zero 1 2 3 4",
	"5 6 7 8 9",
	"zero 1 2 3 4 5 6 7 8 9",
	"a good chuckle",
	"sing 'n shout again",
	"embrace no further",
	"adam's jolly decision",
	"happy new year",

	// then, prompts that might characterize your dialogs:
	// (change to fit your problem)
	"9 8 7 6 5",
	"4 3 2 1 zero",
	"mary had a little lamb",
	"it's fleece was white as snow",
	"everywhere that mary went",
	"the lamb was sure to go",
	"twinkle twinkle little star",
	"how i wonder what you are",
	"up above the world so high",
	"like a diamond in the sky",
};

const char* saveWv[] =
{
	// first, prompts that cover the english phoneme set (don't touch!)
	"p0.wav",
	"p1.wav",
	"p2.wav",
	"p3.wav",
	"p4.wav",
	"p5.wav",
	"p6.wav",
	"p7.wav",
	"p8.wav",
	"p9.wav",
	"p10.wav",
	"p11.wav",
	"p12.wav",
	"p13.wav",
	"p14.wav",
	"p15.wav",
	"p16.wav",
	"p17.wav",
};


#define maxCutt		8	// the phoneme construction cutoff
#define maxTutt		int(sizeof(prompts)/sizeof(char*))

Utterance* utts[maxTutt];


////////////////////////////////////////////////////////////////
// The keyboard
//
// We will want our main thread to be able to poll for events
// and if there is no event, to move on.  The simple getc() of
// unix waits for the next char to occur.  We will need to
// create a thread that can park on getc(), when it gets a char,
// puts it in a queue.  That way the main thread can check the
// queue, process any waiting chars, or move on if there are
// none.


#define KBD_Length	128		// must be power of 2
#define NO_KEYS_AVAILABLE	0xfff

class CharFifo
{
 protected:
	uns8 keyboard[KBD_Length];
	short head, tail;

 public:
	CharFifo() { head=tail=0; }

	// put in at the head, take out at the tail;
	short pop() {
		if( tail==head ) return NO_KEYS_AVAILABLE;
		uns16 rtn = keyboard[tail++];
		tail &= KBD_Length - 1;
		return short( rtn );
	}

	boole push( uns8 key ) {
		short nextHead = (head+1) & (KBD_Length-1);
		if( nextHead == tail )
			return 0;
		keyboard[head] = key;
		head = nextHead;
		return 1;
	}
};

CharFifo kbd;


int keyBoardActive = 1;

#ifdef WIN32
DWORD WINAPI getNextKey( CharFifo* kbd )
{
    INPUT_RECORD record;
    DWORD numRead;
	HANDLE stdIn = GetStdHandle(STD_INPUT_HANDLE);

	while( keyBoardActive )
	{
		if( 0 == ReadConsoleInput( stdIn, &record, 1, &numRead ) )
			return 0; // error situation...

		if( record.EventType != KEY_EVENT )
			continue;		// don't care about other console events

		if( record.Event.KeyEvent.bKeyDown )
		{
			char c = record.Event.KeyEvent.uChar.AsciiChar;
			kbd->push( c );
		}
    }
	return 0;
}

#else
void* getNextKey( void* kbd )
{
	while( keyBoardActive )
	{
		uns8 c = (uns8 )getc( stdin );
		((CharFifo* )kbd)->push(c);
		if( c == 'x' || c == 'X' )
		Sleep(2);
	}
	return 0;
}

#endif


////////////////////////////////////////////////////////////////
// Globals 
int uttI=0, lastI=-1, pitch=-1, doPrint;
Utterance *lastUtt = 0;
User *user, *enrolUser;

char resultBuf[2048];
char* resultStr[8];
int resultCnt;


////////////////////////////////////////////////////////////////
// Loops...
//
// The main() will have 2 main loops, a normal recognition loop,
// and the enrollment loop.



short mainPollKbd()
{
	// this function will be called in the 'normal' polling loop to
	// check for keyboard activity. We only print a message the
	// after a processing an utterance.
	if( doPrint )
	{
#ifdef WIN32
		printf( "\nType 'x' exit, 'e' to enroll, 's' save, or a number to adapt the lastUtt;\nsay something like 'apples are red'\n" );
#else
		printf( "\nType 'x' exit, 'e' to enroll, 's' save, or a number to adapt the lastUtt (then enter);\nsay something like 'apples are red'\n" );
#endif
		fflush( stdout );

		doPrint = 0;
	}

	for(;;)
	{
		// process keys
		short key = kbd.pop();
		switch( key )
		{
		 case NO_KEYS_AVAILABLE:
			return 0;

		 default:
			// drain non action keys
			continue;

		 case 's':
		 case 'S':
			if( hhsAdaptHappened( user ) )
			{
				hhsSaveUser( user, "test.usr" );
				printf( "\nSaved adpated models to test.usr\n" );
			}
			else
				printf( "\nUser Models unchanged\n" );
			continue;

		 case 'e':
		 case 'E':
			return 'e';

		 case 'x':
		 case 'X':
			return 'x';

		 case 'p':
		 case 'P':
			if( lastUtt )
				hhsPlayUtt( lastUtt, BARGE_IN );
			continue;

		 case '0':
		 case '1':
		 case '2':
		 case '3':
		 case '4':
		 case '5':
		 case '6':
		 case '7':
		 case '8':
		 case '9':
			key -= '0';
			if( lastUtt && key < resultCnt && hhsAdapt( user, lastUtt, resultStr[key], 0 ) )
			{
				printf( "Adapted! '%s'\n", resultStr[key] );
				fflush( stdout );
				hhsDeleteUtt( lastUtt );
				lastUtt = 0;
			}
			continue;
		}
	}
}


short enrolPollKbd()
{
	// this function is being called in the enrollment polling loop to check
	// for keyboard activity. We only want to print a message when doPrint
	// is signaled.
	if( doPrint )
	{
#ifdef WIN32
		printf( "\nType 'x' to exit, 'u' to undo -- goto previous prompts, 'p' to play the prompt;\nsay: %s\n", prompts[uttI] );
#else
		printf( "\nType 'x' to exit, 'u' to undo -- goto previous prompts, 'p' to play the prompt (then enter);\nsay: %s\n", prompts[uttI] );
#endif

		fflush( stdout );
		//hhsPlayWavFile( saveWv[uttI], 0, BARGE_IN );

		doPrint=0;
	}

	while( uttI < maxTutt )
	{
		short key = kbd.pop();
		switch( key )
		{
		 case NO_KEYS_AVAILABLE:
			return 0;

		 default:
			// drain non action keys
			continue;

		 case 'x':
		 case 'X':
			return 'x';

		 case 'p':
		 case 'P':
			hhsPlayWavFile( saveWv[uttI], 0, BARGE_IN );
			continue;

		 case 'u':
		 case 'U':
			if( uttI > 0 )
			{
				--uttI;
				hhsDeleteUtt( utts[ uttI ] );
				utts[ uttI ] = 0;
				doPrint=1;
			}
			continue;
		}
	}
	return 0;
}


short enrolPitchPollKbd()
{
	// this function is being called in from the enrollment loop just after
	// measuring the pitch.  It checks for keyboard activity and allows the
	// user to adjust the pitch, we give the user the opportunity to
	// manually select the pitch...  We only want to print a message when
	// doPrint is signaled.
	if( doPrint )
	{
		printf( "\nPitch is %d, to changed it enter a number between 45 and 85, or", pitch );
#ifdef WIN32
		printf( "\nType 'x' to exit, 'u' to undo, 'p' to play the prompt;\nsay: %s\n", prompts[uttI] );
#else
		printf( "\nType 'x' to exit, 'u' to undo -- goto previous prompts, 'p' to play the prompt (then enter);\nsay: %s\n", prompts[uttI] );
#endif

		fflush( stdout );
		//hhsPlayWavFile( saveWv[uttI], 0, BARGE_IN );

		lastI = uttI;
		doPrint = 0;
	}

	while( uttI < maxTutt )
	{
		short key = kbd.pop();
		switch( key )
		{
		 case NO_KEYS_AVAILABLE:
			return 0;

		 default:
			// drain non action keys
			continue;

		 case 'x':
		 case 'X':
			return 'x';

		 case 'p':
		 case 'P':
			hhsPlayWavFile( saveWv[uttI], 0, BARGE_IN );
			continue;

		 case 'u':
		 case 'U':
			if( uttI > 0 )
			{
				--uttI;
				hhsDeleteUtt( utts[ uttI ] );
				utts[ uttI ] = 0;
				doPrint=1;
			}
			continue;

		 case '0':
		 case '1':
		 case '2':
		 case '3':
		 case '4':
		 case '5':
		 case '6':
		 case '7':
		 case '8':
		 case '9':
			printf( "%c", key );
			fflush( stdout );
			key -= '0';
			pitch = pitch * 10 + key;
			if( pitch > 85 ) {
				pitch = key;
				printf( "\n%d", key );
			}
			continue;
		}
	}
	return 0;
}




////////////////////////////////////////////////////////////////////////
// main()


int main( int argc, char** argv )
{
	// initialize the sdk.
	if( 0 == hhsNewGlobalRecognizer( "\t3673:0504:29742" ) )
	{
		printf( "Failed initializing!\n" );
		return 0;
	}

	// setup the keyboard polling
	// set the log folder to some reasonable place
#ifdef WIN32
	CreateThread( 0, 0, (LPTHREAD_START_ROUTINE )getNextKey, (void* )&kbd, 0, 0 );
#else
	pthread_t kbdThread;
	pthread_create( &kbdThread, 0, getNextKey, &kbd );
#endif
	hhsSetLogDir( "./" );

	// lang = english -- is actually the default (here for completeness)
	hhsSetLanguageId( ENGLISH_LANGID );

	// set an upper limit on how much memory to use during
	// recognition; this use to be an important call when
	// running in an embedded environment.
	hhsSetMaxAllocDuringRecognition( 50*1024*1024 );

	// load the .voc first so the word pronunciations exist
	// then load .rul, the words in the rules will be found
	// (not guessed)
	if( 0 == hhsLoadWords( "test.voc" ) ||
		0 == hhsLoadRules( "test.rul" ) )
	{
		printf("Can't open .voc or .rul file\nCheck the hhsLog.txt file");
		// failure!!!
		return 0;
	}
	hhsSaveRules( "test1.rul" );  // just testing
	hhsSaveWords( "test1.voc", 1 );  // 1 -> only rule words

	// allocate the memory for the models
	user = hhsNewUser();
	void* userRules[10];	// usr-rul, usr-rul, usr-rul, etc
	// even entries are our user models

	userRules[0] = userRules[2] =userRules[4] =userRules[6] = userRules[8] = (void* )user;
	// notice we have not yet loaded the user models -- we have
	// only allocated that memory, but that address won't change
	// once we get around to initializing it with real data --
	// so alright to use in our usr-rul pairs.

	// odd entries are RULEs
	// you can open test.rul with an editor to 'inspect'
	// the RULEs Fruit, GarbageRule are there.
	userRules[1] = hhsGetRule( "Fruit" );
	userRules[3] = hhsGetRule( "GarbageRule" );
	// maybe more rule later...

	// the usr-rul pair -> the rule will be recognized with the
	// specified usr models...  You could have multiple models
	// (for people with different voices running many copies of
	// diffenet rules --> effectively speaker independence or
	// good recognition for a known set of speakers) (caution!
	// you must not use the same rule twice, make a copy with a
	// differnt name)

	// set up the 'Fruit' state with the 2 usr-rul pairs.
	hhsSetUserRulesOfState( "Fruit", userRules, 4 );

	// cap the length of the choice list to 8
	hhsSetMaxResultCount( lengthof(resultStr) );

	short status=0, key=0;
	int i;

	// main loop
	Main_Loop:
	if( 0 == hhsLoadUser( user, "test.usr" ) )
		goto Enroll_Loop;

	// this sets the default signal processing to match our
	// test.usr models.  If the signal processing is different
	// from our models, the sdk will just have to do more work
	// in the recognition routine; It is more efficient to have
	// the correct work done up front.
	hhsSetSignalProcessing( user, 1 );

	// set the 'Fruit' state to be recognized.
	hhsSetRecogState( "Fruit" );

	// this sets the number of frames need to be seen at the end
	// of recognition to believe that the speaker is really done
	// talking. The bigger the number the longer the wait after
	// the end of utterance before an answer will pop out.  Too
	// small a number, if the speaker pauses a little between
	// words, the recognizer might be signaled that the speaker
	// is done at the pause... Not good.  Though the earlyOut
	// condition will be in effect for the rules 'Fruit' (the
	// 'GarbageRule' has a negative penalty the flag for no
	// earlyOut);
	hhsSetNumFramesOfTerminalSilence( 50 );
	// earlyOut condition occurs when the terminal silence at
	// the end of a rule is the very best scoring node of all
	// the active nodes and 20 frames of silence have gone by.
	// Then the recognizers quits, it does not wait the full 4o
	// frames.  So by setting
	// hhsSetNumFramesOfTerminalSilence(50) -- if the terminal
	// silence is not that best scoring place in the rules, then
	// we wait for 50 frames of silence.  But if the very best
	// scoring spot is a valid answer we finish after only
	// waiting 20 frames.  Crisp response when we 'believe' we
	// have a good answer.


	// pruneThresh controls how vigorously we are pruning...  We
	// are scoring nodes in our rule/word network.  At any given
	// moment there is set of 'good' scoring nodes.  In fact
	// there is a 'best' scoring node.  pruneThresh=300 -> if a
	// node has a score more than 300 points from the best
	// score, we prune it.  We quit scoring it -- quit letting
	// it seed its following nodes.  Setting pruneThresh lower
	// means we prune more aggressively (do less work, use less
	// memory) but we risk pruning the right answer should it
	// initially score poorly (there is a background noise, the
	// speaker mumbles a little) when with more speech the right
	// answer would start fitting again and 'win' in the end...
	// Recommendation -- it's best not to prune too hard.
	hhsSetPruneThresh( 300 );


	// this turns on our noise strategy -- read about it in the hhssdk.txt
	hhsSetNoiseInjectionThresh(1);

	// initalize some global variable
	resultCnt = 0;
	if( lastUtt ) // if there is an utt hanging around...
	{
		hhsDeleteUtt( lastUtt );
		lastUtt = 0;
	}
	doPrint=1;

	for(;;)
	{
		// poll the kbd
		key = mainPollKbd();
		if( key == 'e' )
			goto Enroll_Loop;

		else if( key == 'x' )
			goto CmdRec_Exit;

		// check if there is an utt...
		Utterance *utt = hhsListenDirectPolled( LstnCONTINUED|LstnKEEPWAVDATA|LstnKEEPEMODATA, &status );
		// (Listen() turns the mic on if it was off)

		if( utt )
		{
			// got one! when we have an utterance, status holds the result
			// count. We know the count is at least 1, otherwise we wouldn't
			// have gotten an utterance back; while we think about it we
			// will need to print next time through, set doPrint.
			doPrint = 1;

			unsigned short lam[3];
			static short lam0 = 0;
			lam[0]=hhsGetEmoLambdaInfo( utt, lam+1, lengthof(lam)-1 );
			if( lam[0] > 0 )
			{
				// I like to change the scale a litte for lam[1] and lam[2]
				lam[1] >>= 2;
				lam[2] >>= 3;
				printf( "Avg Pitch WaveLength: %d  Std Dev: %d   Fitting Stress: %d\n", lam[0], lam[1], lam[2] );
				if( lam0 == 0 )		lam0 = lam[0];
				// this emo lambda info will be used for rejection...
			}
			
			unsigned semid[32];
			resultCnt = hhsResultsToFirstSemanticIds( utt, semid, lengthof(semid) );
			// hhsResultsToFirstSemanticIds() gets the first semantic id of
			// each result; for example, if there were 4 resuls than there
			// will be 4 items coming back from the function call. The first
			// semantic id is always the semantic id associated with the
			// rule that is expressed in the result.  Referring back to
			// test.rul, 'Fruit' -> 100  and 'GarbageRule' -> 666   (btw if we
			// check, resultCnt will equal status)

			if( semid[0] == 666 )
			{
				// garbage rule is top choice -- reject this and bail...
				printf( "got garbage -- ignoring\n" );
				hhsRejectUtt( utt );
				continue;
			}
			if( lam[0] && (lam[1] > 300 || lam[2] > 300) )
			{
				// the pitch standard deviation is big, or the fitting stress is
				// big -- reject this and bail...  In your environment (your
				// background noise) you should look at the lam[] rejection
				// criteria				
				printf( "got env noise -- ignoring\n" );
				hhsRejectUtt( utt );
				continue;
			}
			
			int score[32];
			resultCnt = hhsResultsToScores( utt, score, lengthof(score) );
			// this brings back all the hypothesis scores for the
			// choices on the choice list.

			// This is my favorite way to get the ascii results.
			// You hand it a big buf and all the results come
			// back at once.
			if( hhsResultsToAscii( utt, 0, resultBuf, lengthof(resultBuf) ) )
			{
				// hhsResultsToAscii returns everything in resultBuf;
				// beginning of resultBuf is an array of shorts that are the
				// offsets to the ascii result strings.  Following that are
				// the strings themselves.  The strings will be null
				// terminated.

				// coerce the start of the chars to a pointer to shorts
				short *offset = (short* )resultBuf;
				
				for( i=0; i<resultCnt; ++i )
				{
					// If the resultsBuf wasn't big enough for all the
					// strings then the offset will be 0 for the missing
					// strings.  For example suppose you had 8 results and
					// but resultBuf was only big enough to return the first
					// 5, then offset[] will look somthing like:
					// offset[0] = 16, offset[1] = 60, offset[2] =
					// 110, offset[3] = 185, offset[4] = 225,
					// offset[5] = 0, offset[6] = 0, offset[7] = 0;
					// We could then call again:
					// hhsResultsToAscii( utt, 5, resultBuf, lengthof(resultBuf) );
					// and offset will have 0, in offset[0 thru 4] and
					// have nozero offsets starting in offset[5]...

					// but we are using offset[1024] -- huge and who cares
					// if the last couple results are not shown...
					if( semid[i] == 666 || offset[i] == 0 )
					{
						// this answer is garbage, or we have lots of
						// results and we don't want to screw up the
						// addresses recorded in resultStr[]... so bail
						// after truncating resultCnt.
						printf( "%d) garbage rule\t\t\tdelta: %d\n", i, score[i] - score[0] );
						resultCnt = i;
						break;
					}
						
					// capture the string address, we might use it later if
					// the person adapts using this utterance.
					resultStr[i] = resultBuf + offset[i];
					if( i == 0 )
					{
						// display semantic ids
						int fruit=0, color=0;
						unsigned reselem[120];
						ResultElement *resElem = (ResultElement* )reselem;
						int cnt = hhsIthResultElements( utt, 0, reselem, lengthof(reselem) );
						cnt = cnt*sizeof(unsigned)/sizeof(ResultElement);
						for( int j=0; j<cnt; ++j )
						{
							if( resElem[j].type == RE_WORD &&
								resElem[j].semanticId != UNUSED_SEMANTIC_ID )
							{
								if( fruit==0 )
									fruit = resElem[j].semanticId;
								else
								{
									color = resElem[j].semanticId;
									break;
								}
							}
						}
						printf( "%d) %s\t\ttotal: %d\tsemids: (%d, %d)\n", i, resultStr[i], score[0], fruit, color );
					}
					else
						printf( "%d) %s\t\tdelta: %d\n", i, resultStr[i], score[i] - score[0] );

				}

				if( lam[0] )
				{
					// here is a loose look at the changing wavelength
					int del = lam0 - lam[0];
					if( del < 0 ) del = -del;

					if( (del<<3) > lam0 )
					{
						// seems like lam[0] has changed from utt to utt
						// figure up to 12% variation is not unusual...
						// maybe should have rejected this utt...
						// only track lightly
						lam0 = (3*lam0 + lam[0])>>2;
					}
					else
					{
						// normal vaiation
						lam0 = (lam[0] + lam0) >> 1;
						hhsSetBestInputGain( utt );
					}
				}

				// keep the last utt around so that the user can adapt if he
				// wants to.  So out with the old, in with the new.
				if( lastUtt )
					hhsDeleteUtt( lastUtt );
				lastUtt = utt;

				continue;
			}
			else
			{
				// shouldn't get here...
				printf( "no result\n" );
				hhsRejectUtt( utt );
			}
		}
		else
		{
			// no utt, make sure status is copacetic
			if( status < 0 )	goto CmdRec_Exit;

			// wait a bit for new sound data.
			Sleep( 20 );
		}
	}


	// enroll loop
	Enroll_Loop:

	if( 0 == hhsLoadUser( user, "newuser.usr" ) )
	{
		printf( "Missing newuser.usr models! Exiting\n" );
		return 1;
	}
	hhsSetSignalProcessing( user, 1 );

	// turn off the 'Fruit' state
	hhsSetRecogState( "" );
	// setting the empty string as the recognition state ->
	// means no named state recognition.
	
	hhsSetNumFramesOfTerminalSilence( 60 );
	hhsSetPruneThresh( 320 );
	hhsSetNoiseInjectionThresh(0);  // turn off noise strategy
	if( lastUtt )
	{
		hhsDeleteUtt( lastUtt );
		lastUtt = 0;
	}
	printf( "Begin Enrollment\n\nOnly enroll in a quiet evironment\n\n" );
	doPrint=1;

	while( uttI<maxTutt )
	{
		// poll the kbd
		key = ( pitch>0  ? enrolPitchPollKbd() : enrolPollKbd() );

		if( key == 'x' )
			goto Main_Loop;

		Utterance *utt = hhsListenDirectPolled( LstnCONTINUED|LstnKEEPWAVDATA|LstnCOMPLETE|LstnMODAL, &status );
		// the key flag is LstnCOMPLETE, that means we collect the utterance
		// until it is complete and we don't care about if there is recognition.
		// The LstnMODAL flag means that if there were any other apps waiting
		// for utterances, they don't get the next one, It will go to this app.
		// (Listen() turns the mic on if it was off)
		
		if( utt )
		{
			doPrint = 1;
			hhsSetPrompt( utt, prompts[uttI] );
			// hhsSetPrompt() stores the string in the utt so that later
			// when the utt is used to adapt the models, the words of the
			// utterance are known.

			//hhsWriteWavFileUtt( utt, saveWv[uttI] );
			utts[ uttI++ ] = utt;

			// use the first couple utterances are collected to set the
			// correct gain
			if( uttI<3 )
				hhsSetBestInputGain( utt );

			else if( uttI == 3 )
			{				
				// the 3rd utt is "0 1 2 3 4 5 6 7 8 9" is nice and long and
				// is used to measure the person's pitch -- relatively large
				// cpu requirement.  We stop listening to prevent the
				// accidental collection of an utterance while doing this
				// work. 
				hhsStopListening();
				printf( "listening paused.....while busy\n" );
				fflush( stdout );

				// load the newuser.usr models -- these are our base models
				// with the most data behind them.
				if( enrolUser == 0 )
					enrolUser = hhsNewUser();

				if( 0 == hhsLoadUser( enrolUser, "newuser.usr" ) )
				{
					// something went wrong!  undo...
					printf( "failed opening newuser.usr\n" );
Bail_3rd_Utt:
					hhsDeleteUtt( utt );
					utts[ --uttI ] = 0;
					pitch=-1;
					continue;
				}
				// the newuser.usr models are double models. Collapse the
				// models to the simpler single models.  The simpler single
				// models are easier to adapt to the speakers voice (see
				// hhssdk.txt for a discusion)
				hhsSingleModels( enrolUser );

				// figure what pitch/signal processing best fits the
				// speakers voice.
				User *usr = hhsPickUserAndPitch( utt, enrolUser );

				if( usr == 0 )	// try again with a new utt
				{
					printf( "pickUserArray failed\n" );
					goto Bail_3rd_Utt;
				}

				// if hhsPickUserAndPitch() was successful, it may have
				// changed the pitch measure embedded in the models to
				// better match the speaker's voice.  Get the pitch.
				pitch = hhsGetPitch( enrolUser );

				// it must be a legal pitch
				if( pitch < 44 || pitch > 86 )
					goto Bail_3rd_Utt;

				lastI = 0;

				// we have a 'good' pitch measure, we test some other
				// models for a better fit at this pitch.
				const char* otherUsr[] = { "hm.usr", "jc.usr", "dm.usr" };
				User* pickUserArray[4];
				assert( lengthof(pickUserArray) == lengthof(otherUsr)+1 );

				for( i=0; i<lengthof(otherUsr); ++i )
				{
					pickUserArray[i] = hhsNewUser();
					if( 0 == hhsLoadUser( pickUserArray[i], otherUsr[i] ) )
					{
						assert( 1 == 0 );
					};
					// force our simplest type of models
					hhsSingleModels( pickUserArray[i] );
				}
				pickUserArray[ i++ ] = enrolUser;

				// we ease up on pruning
				hhsSetPruneThresh( 360 );
				// pick the best user at this pitch
				usr = hhsPickUser( utt, pickUserArray, i, pitch );
				hhsSetPruneThresh( 320 );

				if( usr == 0 )
				{
					--i; // keep enrolUser (it was loaded last)
					while( i>0 )
						hhsDeleteUser( pickUserArray[--i] );

					goto Bail_3rd_Utt;
				}

				while( i>0 )
				{
					if( usr != pickUserArray[--i] )
						hhsDeleteUser(  pickUserArray[i] );
				}
				pitch = hhsGetPitch( (enrolUser=usr) );
				continue;
			}

			else if( uttI == 4 )
			{
				// the user may have manually changed the pitch...
				// if not hhsSetPitch() has no effect, 
				if( pitch > 44 && pitch < 86 )
					hhsSetPitch( enrolUser, pitch );

				hhsSetSignalProcessing( enrolUser );
				printf( "\nUsing pitch: %d\n", hhsGetPitch( enrolUser ) );
				pitch = -1;
			}
			else if(  uttI == maxCutt )
			{
				// We have collected at least one version of every phoneme
				// type, time to do our first major massage on the
				// modele (C stands for construct).
				hhsStopListening();
				printf( "listening paused.....\n" );
				fflush( stdout );

				// adapt(): learn this initial data
				// first do batchAdapt() ->
				// it redoes signal processing
				hhsBatchAdapt( enrolUser, utts, maxCutt, 1 );
				for( int j=4; j>=0; --j )
				{
					// skipping the first 2 utterances... new users
					// unfamiliar with enrollment have a tendency to screw
					// up the first utterance or two... safest to just
					// ignore them.
					for( i=2; i<maxCutt; ++i )
						hhsAdapt( enrolUser, utts[i], 0, j & 1 );
				}

				// make a stab creating the full models set.
				// hhsUpdateUserCousins() uses an adhoc algorithm to guess
				// phoneme models based on very little data.
				hhsUpdateUserCousins( enrolUser );
			}

			else if( uttI == maxTutt )
			{
				// time to exit, do finishing adaptation
				hhsStopListening();
				printf( "listening paused.....\n" );
				fflush( stdout );

				// passing through the data 3 times
				for( int j=2; j>=0; --j )
				{
					for( i=2; i<maxTutt; ++i )
						hhsAdapt( enrolUser, utts[i], 0, j & 0x1 );
				}
				// our last pass -- accumualte all the delta before changing
				// the single models.
				for( i=2; i<maxTutt; ++i )
					hhsAdaptAccumulate( enrolUser, utts[i], 0, 1 );

				hhsAdaptFinish( enrolUser, 1 );

				// double the number of models...
				hhsDoubleModels( enrolUser );

				// do the same regime with our new double models
				for( int j=2; j>=0; --j )
				{
					for( i=2; i<maxTutt; ++i )
						hhsAdapt( enrolUser, utts[i], 0, j & 0x1 );
				}

				// accumualte all the delta before changing the double
				// models.
				for( i=2; i<maxTutt; ++i )
					hhsAdaptAccumulate( enrolUser, utts[i], 0, 1 );

				hhsAdaptFinish( enrolUser, 1 );

				// here is our attempt: make the best models we can with
				// adapted double models (adhoc algorithm)
				hhsUpdateUserCousins( enrolUser );

				hhsSaveUser( enrolUser, "test.usr" );

				// clean up
				for( i=0; i<maxTutt; ++i )
					if( utts[i] ) hhsDeleteUtt( utts[i] );

				memset( utts, 0, sizeof(utts) );

				goto Main_Loop;
			}
		}
		else
		{
			if( status < 0 )	goto CmdRec_Exit;
			Sleep( 20 );
		}
	}


CmdRec_Exit:
	hhsStopListening();

	if( status < 0 )
		printf( "Exit with bad status %d\n", status );

	if( lastUtt )
		hhsDeleteUtt( lastUtt );

	hhsDeleteUser( user );
	if( enrolUser )
		hhsDeleteUser( enrolUser );

	hhsDeleteGlobalRecognizer();

	return 0;
}
